'=========================================================================
' Title:        Profibus DP SPC3 Driver for P297 Daughter Board
' Module:       P297DR106.BAS
' Platform:     MC Series 2xx + P297 Daughter Board
'-------------------------------------------------------------------------
' Revision:     1.06
' Author:       Chris Backhouse + Geoff Dixon
' Rev Date:     05 May 2004
'-------------------------------------------------------------------------
'             Copyright (c) 2002-2004 Trio Motion Technology Ltd
'
'                      Trio Motion Technology Ltd.
'                Shannon Way, Tewkesbury, Glos. GL20 8ND
'                            United Kingdom
'                          www.triomotion.com
'=========================================================================
' DESCRIPTION:
'  Profibus driver for Cyclic Data Transfer
'  This program sets up the SPC3 chip for transfer of 16 integers from
'  master and 16 intergers to master on a cylcic basis as determined by
'  master unit.
'  Variables:
'   VR(vbase) to VR(vbase+15)    : Data TO master (16bit int)
'   VR(vbase+16) to VR(vbase+31) : Data FROM master (16bit int)
'   VR(vbase+32) = PLC status. 0=running, 2=PLC in configure mode
'   VR(vbase+33) = Connection status
'
' V1.00: 08/01/2001 Beta Release includes full 16+16 VR transfer
' V1.01: 15/03/2001 Reversed byte order for Word transfer
' V1.02: 24/04/2002 Added support for "Leave Data Ex". (S7 config sequence)
'                   Added local timout if no connection after BR detect.
' V1.03: 08/08/2002 Allows VR segment used to be changed by setting a variable
' V1.04: 23/10/2002 Made local timeout value settable.  See "localtimeout"
' V1.05: 25/03/2003 Fixed bug in sending negative numbers.
' V1.06: 05/05/2004 Added a connection status VR

restart:
RESET
node = 5 '    Profibus node address
debug = TRUE 'Set TRUE to get debug messages printed to terminal
db=1 '        Daughter Board slot number
vbase = 20 '  VRs for data transfer
localtimeout = 5000 'time in msec
VR(vbase+33)=0

GOSUB user_dps_reset

TICKS=localtimeout
REPEAT
    GOSUB dps2_ind
    'check local watchdog
    IF TICKS<0 THEN GOTO restart
    'has profbus watchdog timed out?
    IF timeout_flag=TRUE THEN GOTO restart
UNTIL FALSE
STOP


user_dps_reset:


' Hardware reset SPC3:
PROFIBUS(db,1,808,8)
WA(50)
PROFIBUS(db,1,808,0)

' Hardware Watchdog Trigger *******

IF debug THEN PRINT "Initialising SPC3 Registers..."

PROFIBUS(db,1,8,4)'  Set Hardware Mode 1 byte: Force Offline
WA(50)

' Clear SPC3 Memory:

FOR byte=22 TO 1023
    PROFIBUS(db,1,byte,0)
NEXT byte

PROFIBUS(db,1,59,05)'    Set Ident High
PROFIBUS(db,1,58,149)'   Set Ident Low
PROFIBUS(db,1,22,node)'  Set Station Address

' Mode 0 registers must be set in OFFLINE mode:

PROFIBUS(db,1,6,0+0)'    Mode 0 low byte: SYNC and FREEZE mode are OFF
PROFIBUS(db,1,7,1+0+4)'  Mode 0 high byte: DP_MODE on, 10ms ticks
PROFIBUS(db,1,57,255)'   Disallow address change
PROFIBUS(db,1,10,255)'   timeout is 2.55 sec ?

' Define buffers for Dout/Din:

PROFIBUS(db,1,26,32)'    Length of 3 Dout buffers
PROFIBUS(db,1,27,60)'    Segment address of Dout buffer 1
PROFIBUS(db,1,28,64)'    Segment address of Dout buffer 2
PROFIBUS(db,1,29,68)'    Segment address of Dout buffer 3
PROFIBUS(db,1,30,32)'    Length of 3 Din buffers
PROFIBUS(db,1,31,72)'    Segment address of Din buffer 1
PROFIBUS(db,1,32,76)'    Segment address of Din buffer 2
PROFIBUS(db,1,33,80)'    Segment address of Din buffer 3

' Define diagnostic buffers:

PROFIBUS(db,1,36,10)'    Length of Diag buffer 1
PROFIBUS(db,1,37,10)'    Length of Diag buffer 2
PROFIBUS(db,1,38,14)'    Segment address of Diag buffer 1
PROFIBUS(db,1,39,18)'    Segment address of Diag buffer 2
PROFIBUS(db,1,14*8+6,4)' Load buffer with additional length
PROFIBUS(db,1,18*8+6,4)' Load buffer with additional length


' Define Auxiliary buffers:

PROFIBUS(db,1,40,16)'    Length of Aux buffer 1
PROFIBUS(db,1,41,16)'    Length of Aux buffer 2
PROFIBUS(db,1,42,1)'     Control bits for Auxiliary buffers
PROFIBUS(db,1,43,22)'    Segment address of Aux buffer 1
PROFIBUS(db,1,44,24)'    Segment address of Aux buffer 2

' Define Set_Slave_Address buffer:

PROFIBUS(db,1,45,4)'    Length of Set_Slave_Address buffer
PROFIBUS(db,1,46,26)'   Segment address of Set_Slave_Address buffer

' Define Set_Param buffer:

'PROFIBUS(db,1,47,12)'    Length of Set_Param buffer
PROFIBUS(db,1,48,27)'   Segment address of Set_Param buffer

' Define Check_Config buffer:

PROFIBUS(db,1,49,2)'    Length of Check_Config buffer
PROFIBUS(db,1,50,29)'   Segment address of Check_Config buffer

' Define Get_Config buffer:

PROFIBUS(db,1,51,2)'    Length of Get_Config buffer
PROFIBUS(db,1,52,30)'   Segment address of Get_Config buffer

' Pointer to configuration data:
cd_ptr=PROFIBUS(db,0,52)*8
IF debug THEN PRINT "get_cfg_ptr=";cd_ptr[0]

' Write configuration data:
PROFIBUS(db,1,cd_ptr,95)' number of input words:163
PROFIBUS(db,1,cd_ptr+1,111)' number of output words:16

' Set Indication Functions:

PROFIBUS(db,1,4,NOT(2+4+8))'
PROFIBUS(db,1,5,NOT(1+2+4+8+16+32))

' Set user watchdog value:

PROFIBUS(db,1,24,100)
PROFIBUS(db,1,25,100)
PROFIBUS(db,1,8,32)' Reset watchdog timer

' send the first input (input to master) buffer:
GOSUB get_din_ptr

' Fetch the first diagnosis buffer, initialize service bytes:
GOSUB get_diag_buf_ptr

PROFIBUS(db,1,10,255)'   set wd_baud_ctrl_val *****

PROFIBUS(db,1,8,1)'  Set Hardware Mode 1 byte: START_SPC3

RETURN


dps2_ind:' This function is normally called when an interupt occurs

    intregl=PROFIBUS(db,0,0) ' Read Interupt Register Low Byte
    intregh=PROFIBUS(db,0,1) ' Read Interupt Register High Byte
    intr = (intregl + intregh*256)
    IF (intr AND 65519)<>0 THEN TICKS=localtimeout
    ' Set to go offline:
    IF intregl AND 1 THEN
        PROFIBUS(db,1,2,1) ' Acknowledge Interupt
    ENDIF

    ' Go/Leave Data Ex:
    IF intregl AND 2 THEN
        GOSUB go_leave_data_ex_function
        IF res=0 THEN timeout_flag=TRUE
        IF debug THEN PRINT "go/leave data ex = ";res[0]
        PROFIBUS(db,1,2,2) ' Acknowledge Interupt
    ENDIF

    ' Baudrate Detect:
    IF intregl AND 4 THEN
        is_reg=PROFIBUS(db,0,4) AND 48
        IF debug THEN PRINT "is_reg=";is_reg[0]
        IF (is_reg=16) OR (is_reg=32) THEN PROFIBUS(db,1,11,store_mintsdr)
        PROFIBUS(db,1,2,4) ' Acknowledge Interupt
        IF debug THEN PRINT "BR"
    ENDIF

    ' WD_DP_Mode_Timeout:
    IF intregl AND 8 THEN
        GOSUB wd_dp_mode_timeout_function
        IF debug THEN PRINT "WD_DP Mode timeout"
        PROFIBUS(db,1,2,8) ' Acknowledge Interupt
        timeout_flag=TRUE
    ENDIF

    ' User watchdog has timed out:
    IF intregl AND 16 THEN
        PROFIBUS(db,1,2,16) ' Acknowledge Interupt
    ENDIF

    ' New GC Command:
    IF intregh AND 1 THEN
        GOSUB global_ctrl_command_function
        VR(vbase+32)=res
        IF debug THEN PRINT "GC Command = ";res[0]
        PROFIBUS(db,1,3,1) ' Acknowledge Interupt
    ENDIF

    ' New SSA Data:
    IF intregh AND 2 THEN
        PROFIBUS(db,1,3,2) ' Acknowledge Interupt
        IF debug THEN PRINT "New SSA Data = ";res[0]
    ENDIF

    ' New Config Data:
    IF intregh AND 4 THEN
        IF debug THEN PRINT "New Config Data"
        cfg_result = 0
        result = 0
        REPEAT
            'check configuration data until no conflict behaviour:
            cfg_ptr = PROFIBUS(db,0,50)*8' pointer to config data !
            config_data_len = PROFIBUS(db,0,49)
            IF config_data_len <> 2 THEN
                ' In this example there are 16 words in 16 words out
                PROFIBUS(db,1,9,16) ' write EN_CHG_CFG_BUFFER
                cfg_result=PROFIBUS(db,0,17) ' Read user_cfg_data_nok
            ELSE
                ' Length of the config o.k: check bytes
                config0=PROFIBUS(db,0,cfg_ptr)
                config1=PROFIBUS(db,0,cfg_ptr+1)
                IF (cfg_akt0 = config0) AND (cfg_akt1 = config1) THEN
                    result = 0
                    ' Desired config is equal the actual config
                ELSE
                 IF(config0=95 AND config1=111)OR(config0=17 AND config1=33)THEN
                      cfg_akt0 = config0
                      cfg_akt1 = config1
                      result=2
                  ELSE
                      result=1
                  ENDIF
                  IF result = 2 THEN
                      ' Program should acknowledge check_config configuration
                      user_io_data_len_ptr = res
                      PROFIBUS(db,0,16)
                  ENDIF
                ENDIF
            ENDIF
        UNTIL cfg_result<>1
        PROFIBUS(db,1,3,4) ' Acknowledge Interupt
    ENDIF

    ' New Param Data:
    IF intregh AND 8 THEN
        prm_result = 0
        REPEAT
            prm_ptr = PROFIBUS(db,0,48)*8' ??? pointer to data
            param_data_len = PROFIBUS(db,0,47)
            IF debug THEN PRINT "pp=";prm_ptr[0],"  pdl=";param_data_len[0]
            ' Standard param_data_len is 7 bytes
            ' additional bytes are set in COM PROFIBUS (here 5)+7=12
            IF (param_data_len > 12) THEN
                ' Acknowledge not okay:
                b1=PROFIBUS(db,0,prm_ptr+8)
                IF b1 AND(PROFIBUS(db,0,prm_ptr+9)=170) THEN
                    prm_result=PROFIBUS(db,0,15)
                ENDIF
            ELSE
                ' Acknowledge okay:
                prm_result=PROFIBUS(db,0,14)
                IF debug THEN PRINT "p_res=";prm_result[0]
            ENDIF
        UNTIL prm_result<>1

        ' Have received from master standard timeout value:
        store_mintsdr = PROFIBUS(db,0,prm_ptr+3)'for restart
        IF debug THEN PRINT "store_mintsdr=";store_mintsdr[0]
        running=TRUE
        VR(vbase+33)=1

        PROFIBUS(db,1,3,8) ' Acknowledge Interupt
    ENDIF


    ' Diagnostic Buffer Changed:
    IF intregh AND 16 THEN
        PROFIBUS(db,1,3,16) ' Acknowledge Interupt
        IF debug THEN PRINT "diag buf changed"
    ENDIF

    ' Write Read Data Changed:
    IF intregh AND 32 THEN
        PROFIBUS(db,1,3,32) ' Acknowledge Interupt

        ' Handle update to outputs (from master):

        IF running<>0 THEN
          dout_buf_cmd=PROFIBUS(db,0,10)/16 AND 3
          dout_buf_ptr=PROFIBUS(db,0,27+dout_buf_cmd-1)*8
          vrb=vbase+16
          FOR ct=0 TO 15
            hibyte=PROFIBUS(db,0,dout_buf_ptr+(2*ct))
            lowbyte=PROFIBUS(db,0,dout_buf_ptr+(2*ct)+1)
            sint=INT(lowbyte+(hibyte*256))
            IF sint>32767 THEN sint=INT(sint-65536)
            VR(vrb+ct)=sint
          NEXT ct
        ENDIF
        PROFIBUS(db,0,11)' update output buffers

        ' Handle update to inputs (to master):

        din_buf_cmd=PROFIBUS(db,0,8)/16 AND 3
        din_buf_ptr=PROFIBUS(db,0,31+din_buf_cmd-1)*8
        FOR ct=0 TO 15
          sint=VR(vbase+ct)
          IF sint>32767 THEN sint=32767
          IF sint<-32768 THEN sint=-32768
          IF sint<0 THEN sint=sint+65536
          lowbyte=sint AND 255
          hibyte=(sint/256) AND 255

          PROFIBUS(db,1,din_buf_ptr+(2*ct),hibyte)
          PROFIBUS(db,1,din_buf_ptr+(2*ct)+1,lowbyte)
        NEXT ct
        PROFIBUS(db,0,9)' rotate input buffers

    ENDIF

    ' Set EOI in DP_Din_Buffer_State_Machine:
    PROFIBUS(db,1,8,2)
    PROFIBUS(db,1,8,32)' Reset watchdog timer
    IF running THEN loops=loops+1
    IF loops=500 THEN PROFIBUS(db,0,9)

RETURN

wd_dp_mode_timeout_function:
    ' Profibus watchdog has run out.

    user_wd_state = (PROFIBUS(db,0,4)/64) AND 3

    len=6
    IF (PROFIBUS(db,0,12) AND 3)=1 THEN
        PROFIBUS(db,1,36,len)
    ELSE
        IF (PROFIBUS(db,0,12) AND 12)=4 THEN
            PROFIBUS(db,1,37,len)
        ELSE
            len=255
        ENDIF
    ENDIF
    running=FALSE
    VR(vbase+33)=0
    loops=0
    IF debug THEN
     PRINT "Profibus Timeout - Offline"
     PRINT ""
    ENDIF
RETURN

global_ctrl_command_function:
    res=PROFIBUS(db,0,60)
RETURN

go_leave_data_ex_function:
    res=(PROFIBUS(db,0,4)/16) AND 3
RETURN

get_diag_buf_ptr:
   chg_diag_b_new = 0
   chg_diag_b_old = 0

   diag_buffer_sm=PROFIBUS(db,0,12)
   IF (diag_buffer_sm AND 3)=1 THEN
        user_diag_buffer_ptr=PROFIBUS(db,0,38)*8
   ELSE
      IF (diag_buffer_sm AND 12)=4 THEN
         user_diag_buffer_ptr=PROFIBUS(db,0,39)*8
      ELSE
         user_diag_buffer_ptr=0
      ENDIF
   ENDIF
   PROFIBUS(db,0,13)' Make new diag buffer available
   user_diag_flag = TRUE
RETURN


get_din_ptr:
   din_buffer_sm=(PROFIBUS(db,0,8)/16) AND 3
   IF din_buffer_sm THEN
       user_din_ptr=PROFIBUS(db,0,31+(din_buffer_sm-1))*8
   ELSE
       user_din_ptr=0
   ENDIF
RETURN

